;; Generated package description from racket-mode.el  -*- no-byte-compile: t -*-
(define-package "racket-mode" "1.0.20250830.122523" "Racket editing, REPL, and more" '((emacs "25.1") (compat "30.0.2.0")) :commit "f3952cdebb7d92b65bd6755513b51cf01812032a" :authors '(("Greg Hendershott" . "racket-mode-author@greghendershott.com")) :url "https://www.racket-mode.com/")
